@@ -14,7 +14,12 @@ from tornado.iostream import StreamClosedError
 from tornado.websocket import WebSocketHandler, WebSocketClosedError
 
 from jupyter_client.session import Session
-from jupyter_client.jsonutil import date_default, extract_dates
+try:
+    from jupyter_client.jsonutil import json_default, extract_dates
+except ImportError:
+    from jupyter_client.jsonutil import (
+        date_default as json_default, extract_dates
+    )
 from ipython_genutils.py3compat import cast_unicode
 
 from notebook.utils import maybe_future
@@ -40,7 +45,7 @@ def serialize_binary_message(msg):
     # don't modify msg or buffer list in-place
     msg = msg.copy()
     buffers = list(msg.pop('buffers'))
-    bmsg = json.dumps(msg, default=date_default).encode('utf8')
+    bmsg = json.dumps(msg, default=json_default).encode('utf8')
     buffers.insert(0, bmsg)
     nbufs = len(buffers)
     offsets = [4 * (nbufs + 1)]
@@ -88,15 +93,15 @@ class WebSocketMixin(object):
     last_ping = 0
     last_pong = 0
     stream = None
-    
+
     @property
     def ping_interval(self):
         """The interval for websocket keep-alive pings.
-        
+
         Set ws_ping_interval = 0 to disable pings.
         """
         return self.settings.get('ws_ping_interval', WS_PING_INTERVAL)
-    
+
     @property
     def ping_timeout(self):
         """If no ping is received in this many milliseconds,
@@ -109,7 +114,7 @@ class WebSocketMixin(object):
 
     def check_origin(self, origin=None):
         """Check Origin == Host or Access-Control-Allow-Origin.
-        
+
         Tornado >= 4 calls this method automatically, raising 403 if it returns False.
         """
 
@@ -120,18 +125,18 @@ class WebSocketMixin(object):
         host = self.request.headers.get("Host")
         if origin is None:
             origin = self.get_origin()
-        
+
         # If no origin or host header is provided, assume from script
         if origin is None or host is None:
             return True
-        
+
         origin = origin.lower()
         origin_host = urlparse(origin).netloc
-        
+
         # OK if origin matches host
         if origin_host == host:
             return True
-        
+
         # Check CORS headers
         if self.allow_origin:
             allow = self.allow_origin == origin
@@ -193,7 +198,7 @@ class WebSocketMixin(object):
 
 
 class ZMQStreamHandler(WebSocketMixin, WebSocketHandler):
-    
+
     if tornado.version_info < (4,1):
         """Backport send_error from tornado 4.1 to 4.0"""
         def send_error(self, *args, **kwargs):
@@ -206,17 +211,17 @@ class ZMQStreamHandler(WebSocketMixin, WebSocketHandler):
                 # we can close the connection more gracefully.
                 self.stream.close()
 
-    
+
     def _reserialize_reply(self, msg_or_list, channel=None):
         """Reserialize a reply message using JSON.
 
         msg_or_list can be an already-deserialized msg dict or the zmq buffer list.
         If it is the zmq list, it will be deserialized with self.session.
-        
+
         This takes the msg list from the ZMQ socket and serializes the result for the websocket.
         This method should be used by self._on_zmq_reply to build messages that can
         be sent back to the browser.
-        
+
         """
         if isinstance(msg_or_list, dict):
             # already unpacked
@@ -230,7 +235,7 @@ class ZMQStreamHandler(WebSocketMixin, WebSocketHandler):
             buf = serialize_binary_message(msg)
             return buf
         else:
-            smsg = json.dumps(msg, default=date_default)
+            smsg = json.dumps(msg, default=json_default)
             return cast_unicode(smsg)
 
     def _on_zmq_reply(self, stream, msg_list):
