@@ -21,7 +21,7 @@ dummy_date_s = isoformat(dummy_date)
 class DummyMKM(MappingKernelManager):
     """MappingKernelManager interface that doesn't start kernels, for testing"""
     def __init__(self, *args, **kwargs):
-        super(DummyMKM, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
         self.id_letters = iter(u'ABCDEFGHIJK')
 
     def _new_id(self):
@@ -94,7 +94,7 @@ class TestSessionManager(TestCase):
         session = self.create_session(path='/path/to/1/test1.ipynb', kernel_name='python')
         # kill the kernel
         sm.kernel_manager.shutdown_kernel(session['kernel']['id'])
-        with self.assertRaises(KeyError):
+        with self.assertRaises(web.HTTPError):
             self.loop.run_sync(lambda: sm.get_session(session_id=session['id']))
         # no sessions left
         listed = self.loop.run_sync(lambda: sm.list_sessions())
